/*-----------------------------------------------------------------------------
|   AlarmHack
|   
|   A TSR which patches the SndPlaySystemSound API to play the datebook
|   alarm up to 4x.
|
|   This code demonstrates one way to do a TSR on the US Robotics Pilot.
|   It was built using Microsoft Developer Studio using the Macintosh 68K
|   code generator.  It should build with CodeWarrior, but this may involve some
|   tweaking (removing the startup code and all the "Bogus*" references.) 
|
|   The approach is to allocate a fixed chunk of storage RAM, fixup the TSR
|   code and blit the code into the RAM.  My first version of AlarmHack used
|   MemNewPtr and MemPtrSetOwner(0) to allocate memory out of the dynamic heap
|   and then set the owner to the system so the memory wouldn't get freed when
|   AlarmHack stops running.   This worked fine, but with only 32K of dynamic heap 
|   memory available it is incredibly precious so even taking 50-100 bytes 
|   is pretty rude.
|
|   AlarmHack uses a number of Pilot APIs which are marked for System Use only.  
|   This was the only way it could be done.  I hope I don't create too many 
|   headaches for the Palm OS developers when they do the next version of 
|   the Palm OS.  Hopefully they will provide a formal way of doing TSRs.  
|
|   One other approach which I didn't try was to make AlarmHacks TSR code a
|   library using SysLibInstall().  If anyone knows how SysLibs work please
|   let me know.
|
|   An early version of AlarmHack blitted C generated code into the TSR memory.  
|   This version sort of worked, but the code to call the old SndPlaySystemSound 
|   required fixups which were terribly tricky and fragile to maintain.  
|
|	- Wes Cherry
|   (wesc@microsoft.com)
-------------------------------------------------------------WESC------------*/

/* BUGS to fix/todo

appSysLaunchCmdSystemReset -- need to free trap code -- also should
	verify trap address before restoring that too.

clean up code

different alarm sounds (means I have to scrunch the UI....hmmmm)


Wishes for USR:
	An action code for when app is being deleted.


*/


#include <Pilot.h>
#include <SystemPrv.h>		// for SysAppStartup, SysAppExit, SysAppInfoPtr
#include <alarmprv.h>
#include <PilHelp.h>
#include "resource.h"

// Built under msdev
#define MSDEV

typedef void (*PFNSNDPROC)(ULong);

#define kfcCreator 'ALHK'
#define kfcCreatorOld 'alhk'

// Feature mgr indices
#define iftrCode 0
#define iftrOldSndPlaySystemSound 1
#define iftrCrings	2
#define iftrIring 3
#define iftrNags 4
#define iftrCnags 5
#define iftrDsecNag 6
#define iftrAlertShowing 7
#define iftrAlarmStyle 8
#define iftrFullScreenOk 9

// TESTING is defined to make it easier to test -- sets nag delays to be shorter
#define TESTING

#define dsecNagDft 60
#define cnagsMaxDft 5
#define dsecNagInit 0
#define cnagsInfinite 0x7fff
#define cringsDft 1

// This has to be here because of some "as yet unknown requirement of the pilot prc loader" -- Darrin Massena
//char gszBogus[] = "This is a bogus but present string";

char szDb[] = "AlarmHack";

//#define DEBUGPRINT 1
#ifndef DEBUGPRINT
#define DbgPrintL(l, cch, x, y)	 
#endif


int __Startup__()
	{
	SysAppInfoPtr pappi;
	Ptr prevGlobals;					    
	Ptr globalsPtr;

	// Get some startup info from the system (also loads data and sets A5)
	if (SysAppStartup(&pappi, &prevGlobals, &globalsPtr) != 0) 
		{
		// Can't ErrDisplayFileLineMsg here because most likely A5 wasn't
		// set up so we can't get at any strings!
		SndPlaySystemSound(sndError);
		return -1;
		}
	PilotMain(pappi->cmd, pappi->cmdPBP, pappi->launchFlags);
	SysAppExit(pappi, prevGlobals, globalsPtr);
	return 0;
	}



DWord MyFtrGet(DWord featureNum, DWord valueDft)
	{
	DWord dwT;
	if (FtrGet(kfcCreator, featureNum, &dwT) == 0)
		return dwT;
	return valueDft;
	}

Err MyFtrSet(DWord featureNum, DWord newValue)
	{
	return FtrSet(kfcCreator, featureNum, newValue);
	} 

DWord MyFtrGetInit(DWord featureNum, DWord valueDft)
	{
	DWord dwT;
	if (FtrGet(kfcCreator, featureNum, &dwT) == 0)
		return dwT;
	MyFtrSet(featureNum, valueDft);
	return valueDft;
	}


#ifdef DEBUGPRINT
char szT[12];
/*-----------------------------------------------------------------------------
|	itoah
|		
|		integer to ascii, padded to cchPad
|       borrowed from Darrin Massena
-------------------------------------------------------------WESC------------*/
void itoah(char *pszOut, ulong ul, int cchPad)
	{
	char szT[9];
	char *pszT = szT;

	if (cchPad == 0)
		cchPad = 1;

	while (ul != 0) 
		{
		char chDigit = (char)(ul & 0xF) + '0';
		if (chDigit > '9')
			chDigit += 'a' - '9' - 1;
		ul >>= 4;
		*pszT++ = chDigit;
		}

	while (pszT - szT < cchPad)
		*pszT++ = '0';

	while (pszT > szT)
		*pszOut++ = *--pszT;
	*pszOut = 0;
	}

void DbgPrintL(ulong ul, int cchPad, int x, int y)
	{
	itoah(szT, ul, cchPad);
	WinDrawChars(szT, cchPad, x, y);
	}

#endif



#ifdef EXPLORATORY
VOID DisplayVolumes()
	{
	UInt wAlm, wSys, wDef;

	SndGetDefaultVolume(&wAlm, &wSys, &wDef);
	DebugPrintL(wAlm, 4, 0, 20);
	DebugPrintL(wSys, 4, 0, 20);
	DebugPrintL(wDef, 4, 0, 20);
	}
#endif

/*-----------------------------------------------------------------------------
|	PvAllocLockedChunk
|	
|		Allocates a locked chunk out of storage RAM (not dynamic).  
|
|	If no allocation succeeds on the first pass then it makes a second
|	pass compacting the heaps before allocating.
|
|	Free memory allocated by PvAllocLockedChunk via MemFreeChunk.
|	
|	Arguments:
|		int cb:  number of bytes to allocate
|	
|	Returns: pointer to bytes allocated
|	
|	Keywords:
|	
-------------------------------------------------------------WESC------------*/
VoidPtr PvAllocLockedChunk(int cb)
	{
	int ccard;
	int icard;
	int ipass;
	VoidPtr pv;
	
	ccard = MemNumCards();
	for (ipass = 0; ipass < 2; ipass++)
		{
		for (icard = 0; icard < ccard; icard++)	
			{
			// don't bother if this card is all ROM
			if (MemNumRAMHeaps(icard) > 0)
				{
				int idHeap;
				int cidHeap;
				
				cidHeap = MemNumHeaps(icard);
				for (idHeap = 0; idHeap < cidHeap; idHeap++)
					{
					// We don't want ROM or Dynamic heaps
					if (!(MemHeapFlags(idHeap) & memHeapFlagReadOnly) && !MemHeapDynamic(idHeap))
						{
						//UInt cbFree;
						//UInt cbMax;

						// try two -- let's try compacting the heap						
						if (ipass == 1)
							MemHeapCompact(idHeap);
						// MemHeapFreeBytes(idHeap, &cbFree, &cbMax);
						pv = MemChunkNew(idHeap, cb, memNewChunkFlagNonMovable);

						if (pv != NULL)
							{						
							return pv;
							}
						}
					}
				}
			}
		}
	// we couldn't find any heaps so bail
	return NULL;
	}


/*-----------------------------------------------------------------------------
|	MemMoveLockedChunk
|	
|		Blts memory to storage ram.  Takes care to get write access.
|	
|	Arguments:
|		VoidPtr pvDest: Destination, may be storage ram, but this'll work with dynamic ram too
|		VoidPtr pvSrc : source bytes
|		int cb		  : number of bytes to move
|	
|	Returns: pointer to pvDest incremented by cb
-------------------------------------------------------------WESC------------*/
VoidPtr MemMoveLockedChunk(VoidPtr pvDest, VoidPtr pvSrc, int cb)
	{
	// Get Write Access
	MemSemaphoreReserve(TRUE);
	// Blt Code Bytes
	MemMove(pvDest, pvSrc, cb);
	// Release Write Access
	MemSemaphoreRelease(TRUE);
	return (VoidPtr)((byte *) pvDest+cb);
	}



/*-----------------------------------------------------------------------------
|	TSR code to interrupt SndPlaySystemSound and play the datebook alarm
| more than once
-------------------------------------------------------------WESC------------*/

// Init code -- calls default SndPlaySystemSound for all snd* types
Word rgwCodeInit[] =
	{
	0x4e56, 0xffc8,			// link 16,#-56
	//0x102e, 0x0008,			// move.b 8(a6),d0
	//0xb03c, 0x0005,			// cmp.b d0,5
	0x0c2e, 0x0005, 0x0008,	// cmpi.b #5,8(a6)
	0x6700					// be.w SkipOldTrap
	};

Word rgwCode2[] =
	{
	0x0000,
	0x1f2e, 0x0008,			// move.b 8(a6),-(A7)
	0x4eb9, 0xffff, 0xeeee, // jsr oldTrap
	0x544f,					// addq.l #2,A7
	0x102e, 0x0008,			// move.b 8(a6),d0
	0x6000					// bra.w Return
	};

Word rgwCodeNag[] =
	{
	0x0000,			// bne.w NoAlarm (continued)

	// DmGetNextDatabaseByTypeCreator(true, &si, sysFileTApplication, kfcCreator, true, &cardNo, &dbID);
	0x486e, 0xfffc, 		// pea	    _dbID$(a6)
	0x486e, 0xffda, 		// pea	    _cardNo$(a6)
	0x1f3c, 0x0001,			// move.b	1,-(a7)
	0x2f3c, 0x414c, 0x484b, // move.l 	kfcCreator,-(a7)
	0x2f3c, 0x6170, 0x706c, // move.l  sysFileTApplication,-(a7)
	0x486e, 0xffdc,			// pea _si$(a6)
	0x1f3c, 0x0001,			// move.b	1,-(a7)
	0x4e4f, 				// dc.w	    20047	; 4e4fH
	0xa078, 				// dc.w	    41080	; a078H
	0x4fef, 0x0018, 		// lea	    24(a7),a7
	0x4a40,					// tst.w d0
	0x6600,					// bne.w SkipSetAlarm
	};
Word rgwCodeNag2[] =
	{
	0x0000,					// displacement from prev instruction
	// sec = TimGetSeconds();
	0x4e4f, 				// dc.w	    20047	; 4e4fH
	0xa0f5, 				// dc.w	    41205	; a0f5H
	
	//0x2d40, 0xffd4, 		// move.l	    d0,_sec$(a6)
	//0x5a80,					// addq.l	#5,d0
				
	//;  	AlmSetAlarm(cardNo, dbID, 0L, sec, 0);
	0x1f3c, 0x0000,			// move.w	0,-(a7)	// could use clr.b -(a7)
	//0x4227,					// clr.b -(a7)
	//0x2f2e, 0xffd4,  		// move.l	    _sec$(a6),-(a7)
	0x2f00,					// move.l		d0,-(a7)
	0x2f3c, 0x0000, 0x0000, // move.l	0,-(a7)	// could use clr.l -(a7)
	//0x4227,					// clr.b -(a7)
	0x2f2e, 0xfffc,  		// move.l	    _dbID$(a6),-(a7)
	0x3f2e, 0xffda,  		// move.w	    _cardNo(a6),-(a7)
	0x4e4f, 				// dc.w	    20047	; 4e4fH
	0xa238, 				// dc.w	    41528	; a238H
	0x4fef, 0x0010 			// lea	    16(a7),a7	; 00000010H
							// SkipSetAlarm:
							//: NoAlarm
	};
// Cleanup code -- cleans up stack and returns	
Word rgwCodeCleanup[]=
	{
							//: Return
	0x4e5e,					// unlink a6
	0x4e75					// rts
	};

#define cwCode(cring) ((sizeof(rgwCodeInit)+sizeof(rgwCodeCleanup)+sizeof(rgwCode2)+sizeof(rgwCodeNag)+sizeof(rgwCodeNag2))/sizeof(Word))

// fragile #defines
#define iwPatchOldTrapInit 4


// current # of alarm rings
int vcrings = 1;
// current # of nags
int vcnags;
// current dsecNag
int vdsecNag;

#define wCodeVerCur 0
typedef struct _pref
	{
	PFNSNDPROC pfnSndPlaySystemSoundOld;
	byte *pbCode;
	Word wCodeVer;
	Word wUnused;
	DWord dwUnused2;
	} PREF;

/*-----------------------------------------------------------------------------
|	FNukeTrap
|	
|		Removes alhacks trap.
|	
|	Arguments:
|		int crings
-------------------------------------------------------------WESC------------*/

bool FNukeTrap()
	{
	PREF pref;

	// we have to store code and old pointer in pref -- Ftr's go away when you soft reset
	if (PrefGetAppPreferences(kfcCreator, 0, &pref, sizeof(pref)))	
		{
		DbgPrintL((ULong) pref.pbCode, 8, 0, 30);
 		// if old code exists, restore the trap and free the chunk.
		if (pref.pbCode != NULL)
			{
			//pfnSndPlaySystemSound = (PFNSNDPROC) MyFtrGet(iftrOldSndPlaySystemSound, NULL);
			if (pref.pfnSndPlaySystemSoundOld != NULL)
				SysSetTrapAddress(sysTrapSndPlaySystemSound, pref.pfnSndPlaySystemSoundOld);

			MemChunkFree(pref.pbCode);
#ifdef DEBUGPRINT
			WinDrawChars("Alarm Trap nuked!", 17, 0, 60);
			DbgPrintL((DWord) pref.pbCode, 8, 0, 100);
#endif
			pref.pbCode = NULL;
			pref.pfnSndPlaySystemSoundOld = NULL;
			PrefSetAppPreferences(kfcCreator, 0, &pref, sizeof(PREF));
			return true;
			}
		}
	return false;
	}

//Patches SndPlaySystemSound trap, replacing it with our code to ring it more.	
bool FSetTrap()
	{
	PREF pref;

	if (PrefGetAppPreferences(kfcCreator, 0, &pref, sizeof(pref)))	
		{
		if (pref.pbCode)
			{
#ifdef THISWORKSBUTIDRATHERBECONSERVATIVE
			// we already have set the trap and the code is good...just return
			if (pref.wCodeVer == wCodeVerCur)
				return true;
			else
#endif
				FNukeTrap();
			}
		}

	pref.pbCode = (byte *) PvAllocLockedChunk(cwCode(crings)*sizeof(Word));
	if (pref.pbCode == NULL)
		{
		WinDrawChars("Not enough Memory!", 18, 0, 50);
		return false;
		}
	else
		{
		byte *pb;

		pref.wCodeVer = wCodeVerCur;
		pref.pfnSndPlaySystemSoundOld = (PFNSNDPROC) SysGetTrapAddress(sysTrapSndPlaySystemSound);
			
		// patchup our code to point to the old SndPlaySystemSound()
		*(DWord *)&rgwCode2[iwPatchOldTrapInit] = (DWord) pref.pfnSndPlaySystemSoundOld;
		// patch jumps...
		rgwCode2[0] = sizeof(rgwCode2)+2;
		rgwCodeNag[0] = sizeof(rgwCodeNag)+sizeof(rgwCodeNag2);
		rgwCodeNag2[0] = sizeof(rgwCodeNag2);

		pb = (byte *) MemMoveLockedChunk(pref.pbCode, rgwCodeInit, sizeof(rgwCodeInit));
		pb = (byte *) MemMoveLockedChunk(pb, rgwCode2, sizeof(rgwCode2));
		pb = (byte *) MemMoveLockedChunk(pb, rgwCodeNag, sizeof(rgwCodeNag));
		pb = (byte *) MemMoveLockedChunk(pb, rgwCodeNag2, sizeof(rgwCodeNag2));
				
		MemMoveLockedChunk(pb, rgwCodeCleanup, sizeof(rgwCodeCleanup));
		SysSetTrapAddress(sysTrapSndPlaySystemSound, pref.pbCode);
#ifdef DEBUGPRINT
		WinDrawChars("Alarm Trap set!", 15, 0, 50);
#endif
		PrefSetAppPreferences(kfcCreator, 0, &pref, sizeof(PREF));
		}
	return true;
	}


void MySetAlarmDsec(ULong dsec, DWord dwRef)
	{
	UInt cardNo;
	LocalID dbID;
	DmSearchStateType si;
	ULong sec;

	DmGetNextDatabaseByTypeCreator(true, &si, sysFileTApplication, kfcCreator, true, &cardNo, &dbID);
	sec = TimGetSeconds() + dsec;
	AlmSetAlarm(cardNo, dbID, dwRef, sec, 0);
	}

#ifdef FOO
void HorizCenterSz(char *sz, int y)
	{
	SWord dxScreen;
	SWord dyScreen;
	int dx;
	int x;
	int cch;


	cch = StrLen(sz);
	WinGetDisplayExtent(&dxScreen, &dyScreen);
	dx = FntCharsWidth(sz, cch);
	x = (dxScreen-dx)>>1;
	WinDrawChars(sz, cch, x, y);
	}


void DrawNagInfo(int iNag)
	{
	DWord cnags;
	char sz[32];
	char *pch;

//	return;

	cnags = MyFtrGet(iftrCnags, cnagsMaxDft);
	pch = sz;
	*(DWordPtr)pch = 'Alar';
	*(DWordPtr)(pch+4) = 'mHac';
	*(DWordPtr)(pch+8) = 'k Na';
	*(DWordPtr)(pch+12) = 'gs  ';
	if (iNag == 0 || cnags == cnagsInfinite)
		sz[9] = 0;
	else
		{
		pch = sz+15;
		StrIToA(pch, iNag);
		pch += StrLen(pch);
		*pch++ = ' ';
		*pch++ = 'o';
		*pch++ = 'f';
		*pch++ = ' ';

		StrIToA(pch, cnags);
		}
	HorizCenterSz(sz, 130);
	//WinDrawBitmap(kcidbAIB, 10, 125);
	//WinDrawBitmap(kcidbAIB, 120, 125);

	}
#endif*
void HackAlarmAlert(int iNag)
	{
	bool f;
	FormPtr pfrm;
	AlmGlobalsPtr pagt;

	//return ;

	pagt = (AlmGlobalsPtr) GAlmGlobalsP;
	f = pagt->displaying;

	if (!f)
		return;
	if (!MyFtrGet(iftrFullScreenOk, 1))
		return;
					
	// resize the Ok button to be the size of the entire form
	pfrm = FrmGetActiveForm();
	if (pfrm != NULL)
		{
		DWord *pdw;

		pdw = (DWord *) FrmGetTitle(pfrm);
		if (*pdw == 'Remi' && *(pdw+1) == 'nder')
			{
			DWord iRing;

			iRing = MyFtrGet(iftrIring, 0);
	
			// BUG! need to set a flag -- this isn't reliable!
//			if (iNag == 0 && iRing == 0)	// BUG! this should probably be done in DisplayAlarm action code
				{
				ControlPtr pctl;
				RectangleType rc;

				pctl = (ControlPtr) FrmGetObjectPtr(pfrm, 2);
				if (pctl != NULL)
					{
					FrmGetFormBounds(pfrm, &rc);
					if (pctl->bounds.topLeft.x != rc.topLeft.x)
						{
						CtlEraseControl(pctl);
						rc.topLeft.y += 11;
						rc.extent.y -= 11;
						pctl->bounds = rc;
						// Hack! no API to set ctl bounds -- USR is gonna regret this, just like Apple is w/ Copland.  
						CtlDrawControl(pctl);
						}
					}
				}
	//		DrawNagInfo(iNag);
			}
		}
	}


void RemoveOldTrapCode()
	{
	PFNSNDPROC pfnSndPlaySystemSoundOld;
	byte *pbCode;

	if (FtrGet(kfcCreatorOld, iftrCode, &pbCode) == 0) 
		{
		DbgPrintL((ULong) pbCode, 8, 0, 30);
 		// if old code exists, restore the trap and free the chunk.
		if (pbCode != NULL)
			{
			FtrGet(kfcCreatorOld, iftrOldSndPlaySystemSound, &pfnSndPlaySystemSoundOld);
			if (pfnSndPlaySystemSoundOld != NULL)
				SysSetTrapAddress(sysTrapSndPlaySystemSound, pfnSndPlaySystemSoundOld);

			MemChunkFree(pbCode);
			}
		FtrUnregister(kfcCreatorOld, iftrCode);
		FtrUnregister(kfcCreatorOld, iftrOldSndPlaySystemSound);
		}
	}




DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
	{
	EventType evt;
	Word err;

	if (cmd == sysAppLaunchCmdAlarmTriggered)
		{
		int f;
		DWord iNag;
		DWord cnags;
		AlmGlobalsPtr pagt;
		bool fAlertShowingLastTime;
#define psatp ((SysAlarmTriggeredParamType *)cmdPBP)

		//return 0;
		iNag = MyFtrGet(iftrNags, 0);
		cnags = MyFtrGet(iftrCnags, 0);
		pagt = (AlmGlobalsPtr) GAlmGlobalsP;
		f = pagt->displaying;
		if (psatp->ref == 0)
			{
			DWord iRing;
			DWord cRings;

			iRing = MyFtrGet(iftrIring, 0);
			cRings = MyFtrGet(iftrCrings, 0);
			fAlertShowingLastTime = MyFtrGet(iftrAlertShowing, 1);
			if (iRing < cRings && !(iRing > 0 && fAlertShowingLastTime && !f))
				{
			//	if (iNag == 0)
					HackAlarmAlert(iNag);
				iRing++;
#ifdef EXPLORATORY
				// peek for a pen down event -- if there is one then let's stop bugging the user
				EvtGetEvent(&evt, 0);
				if (evt.eType != nilEvent)
					{
					// snerk! is this going to screw everything up since order won't be correct.   I wish USR provided
					// a EvtPeekEvent
					EvtAddEventToQueue(&evt);
					if (evt.eType == penDownEvent)
						goto WaitNag;
					}
#endif
				{
				int isnd;
				void NewAlarmSound(int i);

				isnd = MyFtrGet(iftrAlarmStyle, 0);
				if (isnd == 0)
					{
					PREF pref;

					if (PrefGetAppPreferences(kfcCreator, 0, &pref, sizeof(pref)) && pref.pbCode != NULL)	
						(*pref.pfnSndPlaySystemSoundOld)(sndAlarm<<24);
					}
				else
					{
					NewAlarmSound(isnd-1);
					}	
				}
				// bug! this is a dumb way to kickourselves...might as well loop now...
				SndPlaySystemSound(sndAlarm);
				}
			else
				{
//WaitNag:
				// reset ring count since we are done ringing and are now possibly going to wait for a nag...
				iRing = 0;
				// set wait for nag
				if (f)
					MySetAlarmDsec(MyFtrGet(iftrDsecNag, dsecNagDft), 1);
				}
			MyFtrSet(iftrIring, iRing);
			//psatp->purgeAlarm = true;
			}
		else
			{
			// we are nagging.
			psatp->purgeAlarm = true;

			if (f && iNag++ < cnags)
				{
				HackAlarmAlert(iNag);
				SndPlaySystemSound(sndAlarm);
				}
			else
				iNag = 0;
			MyFtrSet(iftrNags, iNag);
			//psatp->purgeAlarm = true;
			}
		MyFtrSet(iftrAlertShowing, f);
		return 0;
		}

	if (cmd != sysAppLaunchCmdNormalLaunch)
		return 0;
	// cleanup from almhack 0.2
	RemoveOldTrapCode();

	// uninstall warning alert
	if (!FNukeTrap())
		{
		FrmAlert(kidsALT2);
		}

	MyFtrSet(iftrNags, 0);		
	vcnags = MyFtrGetInit(iftrCnags, cnagsMaxDft);
	MyFtrSet(iftrIring, 0);
	vdsecNag = MyFtrGetInit(iftrDsecNag, dsecNagDft);
	vcrings = MyFtrGetInit(iftrCrings, 1);
	FSetTrap();
	FrmGotoForm(kidfForm1);
	do	{
		extern Boolean ApplicationHandleEvent(const EventType *pevt);
		extern Boolean MainFormHandleEvent(const EventType *pevt);

		EvtGetEvent(&evt, evtWaitForever);

		if (!SysHandleEvent(&evt)) 
			if (!MenuHandleEvent(NULL, &evt, &err)) 

			if (!ApplicationHandleEvent(&evt)) 
				if (!MainFormHandleEvent(&evt))
					FrmHandleEvent(FrmGetActiveForm(), &evt);
		} while (evt.eType != appStopEvent);

	return 0;
	}




Boolean ApplicationHandleEvent(const EventType *pevt)
	{
	extern Boolean MainFormHandleEvent(const EventType *pevt);

	if (pevt->eType == frmLoadEvent) 
		{
		ushort fid = pevt->data.frmLoad.formID;
		FormPtr pfrm = FrmInitForm(fid);

		FrmSetActiveForm(pfrm);
		return true;
		}
	return false;
	}

#ifdef EXPLORATORY
int vvol = 0;
void HackSound(int w)
	{
	SndCommandType sct;

	
	sct.cmd = sndCmdFreqDurationAmp << 8;
	switch (w)
		{
	case 1:
		sct.param1 = 500;
		break;
	case 2:
		sct.param1 = 1000;
		break;
	case 3:
		sct.param1 = 2000;
		break;
	case 4:
		sct.param1 = 4000;
		break;

		}
	//sct.param1 = vvol*50+50; //w*100;
	sct.param2 = 200;
	sct.param3 = 64;

	vvol += 1;

	itoah(szT, sct.param1, 4);
	WinDrawChars(szT, 4, 0, 90);

	itoah(szT, sct.param2, 4);
	WinDrawChars(szT, 4, 30, 90);

	itoah(szT, sct.param3, 4);
	WinDrawChars(szT, 4, 80, 90);

	SndDoCmd(NULL, &sct, 1);
	}
#endif



int KcidFromValue(DWord dwVal, short *mpdkcidvalue, int kcidFirst)
	{
	int idkcid;

	for (idkcid = 0; mpdkcidvalue[idkcid] != -2; idkcid++)
		{
		if ((DWord) mpdkcidvalue[idkcid] == dwVal)
			return kcidFirst+idkcid;
		}
	return kcidFirst;
	}



short mpdkciddsec[] =
	{
	-1,
	10,
	30,
	60,
	5*60,
	10*60,
	30*60,
	60*60,
	-2,
	};

#define KcidFromDsec(dsec) (KcidFromValue(dsec, mpdkciddsec, kcidDsecFirst))
#define DsecFromKcid(kcid) (mpdkciddsec[kcid-kcidDsecFirst])

short mpdkcidcnags[] =
	{
	1,
	2,
	5, 
	10,
	20,
	cnagsInfinite
	-2,
	};


#define KcidFromCnags(cnags) (KcidFromValue(cnags, mpdkcidcnags, kcidCnagsFirst))
#define CnagsFromKcid(kcid) (mpdkcidcnags[kcid-kcidCnagsFirst])



#define KcidFromCrings(crings) (kcid1Ring+crings-1)
#define CringsFromKcid(kcid) (((kcid)-kcid1Ring)+1)

#ifdef FOO
int mpchdx[] =
	{
	0,0,0,0,0,0,0,0,  // 0
	0,0,0,0,0,0,0,0,  // 8
	0,0,0,0,0,0,0,0,  // 10
	0,0,0,0,0,0,0,0,  // 18
	2,2,4,8,6,8,7,2,  // 20
	4,4,6,6,3,4,2,5,  // 28
	5,5,5,5,5,5,5,5,  // 30	 // 0..7
	5,5,2,3,6,6,6,5,  // 38   // 8..
	8,5,5,5,6,4,4,6,  // 40
	6,2,4,6,5,8,6,7,  // 48
	5,7,5,5,6,6,6,8,  // 50
	6,6,6,3,5,3,6,4,  // 58
	3,5,5,4,5,5,4,5,  // 60
	5,2,3,5,2,8,5,5,  // 68
	5,5,4,4,4,5,5,6,  // 70
	6,6,4,4,2,4,7,5,  // 78
	5,5,3,8,5,6,6,6,  // 80
	4,11,5,4,8,10,10,10, // 88
	10,3,3,5,5,4,4,7, // 90
	7,10,4,4,8,5,5,6, // 98
	2,2,6,6,8,6,2,5,  // a0
	4,8,5,6,6,4,8,6,  // a8			// (c)..
	5,6,4,4,3,5,6,2,  // b0
	4,2,5,6,8,8,8,5,  // b8
	5,5,5,5,5,5,7,5,  // c0
	4,4,4,4,3,2,3,3,  // c8
	7,6,7,7,7,7,7,5,  // d0
	8,6,6,6,6,6,6,6,  // d8
	5,5,5,5,5,5,8,4,  // e0
	5,5,5,5,2,2,3,3,  // e8
	5,5,5,5,5,5,5,6,  // f0
	7,5,5,5,5,6,5,6};

int vch = 0xf6;
#endif

typedef unsigned char SDC;
#define sdcEnd 0 
#define sdcFreqDur 1
#define sdcEndLoop 2
#define sdcLoop 3
#define sdcDelay 4

#define FreqDur(freq, dur) sdcFreqDur, (SDC) (((freq) & 0xff00) >> 8), (freq) & 0xff, (SDC) (((dur) & 0xff00)>>8), (dur) & 0xff
#define EndLoop() sdcEndLoop
#define Loop(c) sdcLoop, c
#define Delay(dur)  sdcDelay, (dur*sysTicksPerSecond/1000)
#define End() sdcEnd

#ifdef OLD

//__declspec(allocate("_CODE")) 
SDC rgsdc[] =	// siren 0
	{
	Loop(3),
		FreqDur(1000, 200),
		FreqDur(2000, 200),
	EndLoop(),
	End(),
	Loop(3),	// loop 1
		FreqDur(3000, 200),	// sine wave
		FreqDur(4000, 200),
		FreqDur(5000, 200),
		FreqDur(4000, 200),
	EndLoop(),
	End(),
	Loop(3),	// ramp 2
		FreqDur(500, 75),	   // computer (ramp)
		FreqDur(1000, 75),
		FreqDur(2000,75),
		FreqDur(3000, 75),
		FreqDur(4000, 75),
		FreqDur(5000, 75),
	EndLoop(),
	End(),
	Loop(3), // skipalong 3
		FreqDur(1000, 200),
		FreqDur(2000, 500),
	EndLoop(),
	End(),
	Loop(3),	// beethoven
		FreqDur(393*2, 200), // Beethoven's Fuenfste
		Delay(10),
	EndLoop(),
	FreqDur(311*2, 1450),
	Delay(100),
	Loop(3),
		FreqDur(349*2, 200),
		Delay(10),
	EndLoop(),
	FreqDur(294*2, 1450),
	End(),
	Loop(3),	// eurocop
		FreqDur(1000, 750),
		FreqDur(2000, 750),
	EndLoop(),
	End(),
	Loop(24),	// Cricket
		FreqDur(4950, 5),
		FreqDur(5000, 5),
	EndLoop(),
	Delay(500),
	Loop(24),	
		FreqDur(4950, 5),
		FreqDur(5000, 5),
	EndLoop(),
	Delay(500),
	Loop(24),	
		FreqDur(4950, 5),
		FreqDur(5000, 5),
	EndLoop(),
	End(),
	Loop(50),	// bleep
		FreqDur(4996, 2),
		FreqDur(5000, 2),
	EndLoop(),
	End(),
	FreqDur(500, 75),	   // random
	FreqDur(2000,75),
	FreqDur(3000, 75),
	FreqDur(5000, 75),
	FreqDur(500, 75),
	FreqDur(4000, 75),
	FreqDur(1000, 75),
	FreqDur(5000, 75),
	FreqDur(4000, 75),
	FreqDur(500, 75),
	FreqDur(4000, 75),
	FreqDur(3000, 75),
	FreqDur(1000, 75),
	FreqDur(2000,75),
	FreqDur(3000, 75),
	FreqDur(500, 75),
	FreqDur(5000, 75),
	FreqDur(4000, 75),
	FreqDur(1000, 75),
	FreqDur(5000, 75),
	FreqDur(4000, 75),
	FreqDur(500, 75),
	FreqDur(4000, 75),
	End()
	};
#endif 

#ifdef BUG // this doesn't get initialized by the pilot loader!
SDC *rgpsdc[] =
	{
	sdcSiren,
	sdcSine,
	sdcRamp,
	sdcSkipAlong,
	sdcBeethoven,
	sdcEuroCop,
	sdcCricket,
	sdcBleep,
	sdcComputer2
	};
#endif

int WLoadPsdc(SDC *psdc)
	{
	return (*psdc << 8) | *(psdc+1);
	}

void DoAlarmSound(SDC *psdc, int i)
	{
	SDC *psdcLoop;
	int cLoop;
	SndCommandType sct;
	UInt wAlarm;
	UInt wT;
	UInt wTT;

	SndGetDefaultVolume(&wAlarm, &wT, &wTT);
	sct.param3 = wAlarm;

	for(;;)
		{
		switch (*psdc++)
			{
		case sdcEnd:
			if (i-- == 0)
				return;
			break;
		case sdcLoop:
			cLoop = *psdc++;
			psdcLoop = psdc;
			break;
		case sdcEndLoop:
			if (i == 0)
				{
				if (--cLoop > 0)
					psdc = psdcLoop;
				}
			break;
		case sdcFreqDur:
			if (i == 0)
				{
				sct.cmd = sndCmdFreqDurationAmp << 8;
				sct.param1 = WLoadPsdc(psdc);
				sct.param2 = WLoadPsdc(psdc+2);
				SndDoCmd(NULL, &sct, 1);
				}
			psdc += 4;
			break;
		case sdcDelay:
			if (i == 0)
				SysTaskDelay(*psdc);
			psdc ++;
			break;			
			}
		}
	}


void NewAlarmSound(int i)
	{
	DmOpenRef db;

	db = DmOpenDatabaseByTypeCreator(sysFileTApplication, kfcCreator, dmModeReadOnly);
	if (db != 0)
		{
		VoidHand vh;
		vh = DmGet1Resource('tSTR', 3);
		if (vh != NULL)
			{
			Byte *pb;

			pb = (Byte *) MemHandleLock(vh);
			if (pb != NULL)
				{
				DoAlarmSound((SDC *)pb, i);
				MemHandleUnlock(vh);
				}
			DmReleaseResource(vh);
			}
		DmCloseDatabase(db);
		}	
	}

/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  formId - id of the form to display
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			art	2/21/95		Initial Revision
 *
 ***********************************************************************/
static VoidPtr GetObjectPtr (FormPtr pfrm, int objectID)
	{
	return (FrmGetObjectPtr (pfrm, FrmGetObjectIndex (pfrm, objectID)));
	}


Boolean MainFormHandleEvent(const EventType *pevt)
	{
	FormPtr pfrm;
	Word id;

	switch (pevt->eType) 
		{
	case penDownEvent:
		{
#ifdef FOO
		short dx;
		char sz[4];
//		while (vch < sizeof(mpchdx))
			{
			sz[0] = vch;
			sz[1] = 0;
			dx = FntCharWidth(vch);	
//			if (dx != mpchdx[vch])
				{
				DbgPrintL(dx, 4, 0, 40);
				DbgPrintL(vch, 4, 0, 30);
				WinDrawChars(sz, 1, 0, 20);		
//				break;
				}
			vch++;	
			}
#endif
#ifdef DEBUGPRINT
		{
		DWord dwCode;

		FtrGet('alhk', 0, &dwCode);
		DbgPrintL(dwCode, 8, 4, 140);
		}
#endif
//		NewAlarmSound();
		}

		return false;
	case popSelectEvent:
		MyFtrSet(iftrAlarmStyle, pevt->data.popSelect.selection);
		break;
	case ctlSelectEvent:
		id = pevt->data.ctlSelect.controlID;
		if (id >= kcidRingsFirst && id <= kcidRingsLast)
			{
//			NewAlarmSound(id-kcidRingsFirst);
			MyFtrSet(iftrCrings, vcrings = CringsFromKcid(id));
			return true;
			}
		else if (id >= kcidDsecFirst && id <= kcidDsecLast)
			{
			MyFtrSet(iftrDsecNag, vdsecNag = DsecFromKcid(id));
			return true;
			}
		else if (id >= kcidCnagsFirst && id <= kcidCnagsLast)
			{
			MyFtrSet(iftrCnags, vcnags = CnagsFromKcid(id));
			DbgPrintL(id, 4, 0, 80);
			DbgPrintL(CnagsFromKcid(id), 4, 0, 90);
			return true;
			}								
		else if (id == kcidTest1)
			{
			//FSetTrap();
			SndPlaySystemSound(sndAlarm);
			return true;
			}
		else if (id == kcidFullScreenOk)
			{
			MyFtrSet(iftrFullScreenOk, pevt->data.ctlSelect.on);
			}
		break;

	case frmOpenEvent:
		pfrm = FrmGetActiveForm();
		FrmSetControlGroupSelection(pfrm, kgrpRings , KcidFromCrings(vcrings));
		FrmSetControlGroupSelection(pfrm, kgrpDsec , KcidFromDsec(vdsecNag));
		FrmSetControlGroupSelection(pfrm, kgrpCnags , KcidFromCnags(vcnags));
		FrmSetControlValue(pfrm, FrmGetObjectIndex(pfrm, kcidFullScreenOk), MyFtrGet(iftrFullScreenOk, 1));
		// this is quite a rigamorole -- shouldn't the system provide an api for this!
		{
		ListPtr plst;
		ControlPtr pctl;
		int isnd;
		CharPtr pch;
		CharPtr pchLabel;

		isnd = MyFtrGetInit(iftrAlarmStyle, 0);

		plst = (ListPtr) GetObjectPtr (pfrm, 6000);
		pch = LstGetSelectionText(plst, isnd);
		pctl = (ControlPtr) GetObjectPtr(pfrm, 5000);
		pchLabel = CtlGetLabel(pctl);
		StrCopy(pchLabel, pch);
		}		
		FrmDrawForm(pfrm);
		return true;
	case menuEvent:
		switch(pevt->data.menu.itemID)
			{
		case kmidUninstall:
			{
			UInt cardNo;
			LocalID dbID;
			DmSearchStateType si;

			FNukeTrap();
			FrmAlert(kidsALT1);
			if (DmGetNextDatabaseByTypeCreator(true, &si, sysFileTApplication, 'memr', true, &cardNo, &dbID) == 0)
				SysUIAppSwitch(0, dbID, sysAppLaunchCmdNormalLaunch, NULL);
			}
			return true;
			}
		break;
		}
	return false;
	}

// Can't link without this bogus function!
BogusEntry
